@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: GPL-3.0-or-later WITH cryptsetup-OpenSSL-exception */
 /*
     This file is part of sscg.
 
@@ -27,10 +28,9 @@
     version.  If you delete this exception statement from all source
     files in the program, then also delete it here.
 
-    Copyright 2021-2023 by Stephen Gallagher <sgallagh@redhat.com>
+    Copyright 2021-2025 by Stephen Gallagher <sgallagh@redhat.com>
 */
 
-#include <path_utils.h>
 #include <popt.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -43,6 +43,9 @@
 #include "include/dhparams.h"
 
 #include "config.h"
+#ifdef HAVE_GETTEXT
+#include <libintl.h>
+#endif
 
 #define _GNU_SOURCE
 
@@ -65,7 +68,7 @@ get_security_level (void)
 static int
 set_default_options (struct sscg_options *opts)
 {
-  int security_level = get_security_level ();
+  opts->system_security_level = get_security_level ();
 
   opts->ca_mode = SSCG_CERT_DEFAULT_MODE;
   opts->ca_key_mode = SSCG_KEY_DEFAULT_MODE;
@@ -85,48 +88,67 @@ set_default_options (struct sscg_options *opts)
   opts->dhparams_group = talloc_strdup (opts, "ffdhe4096");
   opts->dhparams_generator = 2;
 
+  opts->key_type = talloc_strdup (opts, "rsa");
+
   /* Select the default key strength based on the system security level
    * See:
-   * https://www.openssl.org/docs/man1.1.0/ssl/SSL_CTX_get_security_level.html
+   * https://docs.openssl.org/master/man3/SSL_CTX_set_security_level/#default-callback-behaviour
+   * and
+   * https://docs.openssl.org/master/man7/EVP_PKEY-ML-DSA/#description
    * for the specification of the minimums.
    */
-  switch (security_level)
+
+  /* Security level 2 and below permits lower RSA key-strengths, but SSCG
+  * will set a minimum of 2048 bits and the sha256 hash algorithm.
+  */
+  opts->hash_alg = talloc_strdup (opts, "sha256");
+  opts->rsa_key_strength = 2048;
+
+  /* The minimum EC curve we support is prime256v1. */
+  opts->ec_curve = talloc_strdup (opts, "prime256v1");
+
+  /* We will not support ML-DSA keys below NIST level 2. */
+  opts->mldsa_nist_level = 2;
+
+  switch (opts->system_security_level)
     {
     case 0:
     case 1:
     case 2:
-      /* Security level 2 and below permits lower key-strengths, but SSCG
-       * will set a minimum of 2048 bits and the sha256 hash algorithm.
-       */
-      opts->hash_alg = talloc_strdup (opts, "sha256");
-      opts->key_strength = 2048;
+      /* No change from defaults below case 3 */
       break;
 
     case 3:
-      opts->hash_alg = talloc_strdup (opts, "sha256");
-      opts->key_strength = 3072;
+      opts->rsa_key_strength = 3072;
+      opts->mldsa_nist_level = 3;
+      opts->ec_curve = talloc_strdup (opts, "secp384r1");
       break;
 
     case 4:
       opts->hash_alg = talloc_strdup (opts, "sha384");
-      opts->key_strength = 7680;
+      opts->rsa_key_strength = 7680;
+      opts->mldsa_nist_level = 3;
+      opts->ec_curve = talloc_strdup (opts, "secp384r1");
       break;
 
     default:
       /* Unknown security level. Default to the highest we know about */
-      fprintf (stderr,
-               "Unknown system security level %d. Defaulting to highest-known "
-               "level.\n",
-               security_level);
+      fprintf (
+        stderr,
+        _ ("Unknown system security level %d. Defaulting to highest-known "
+           "level.\n"),
+        opts->system_security_level);
       /* Fall through */
 
     case 5:
       opts->hash_alg = talloc_strdup (opts, "sha512");
-      opts->key_strength = 15360;
+      opts->rsa_key_strength = 15360;
+      opts->mldsa_nist_level = 5;
+      opts->ec_curve = talloc_strdup (opts, "secp521r1");
       break;
     }
 
-  opts->minimum_key_strength = opts->key_strength;
+  opts->minimum_rsa_key_strength = opts->rsa_key_strength;
 
   opts->cipher_alg = talloc_strdup (opts, "aes-256-cbc");
 
@@ -141,8 +163,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
 {
   int ret, sret, opt;
   poptContext pc;
-  char *minimum_key_strength_help = NULL;
+  char *minimum_rsa_key_strength_help = NULL;
   char *named_groups_help = NULL;
+  char *key_type = NULL;
+  char *ec_curve = NULL;
 
   char *country = NULL;
   char *state = NULL;
@@ -151,7 +175,6 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
   char *organizational_unit = NULL;
   char *email = NULL;
   char *hostname = NULL;
-  char *packagename;
   char **alternative_names = NULL;
 
   TALLOC_CTX *tmp_ctx = talloc_new (NULL);
@@ -167,27 +190,26 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
   if (ret != EOK)
     goto done;
 
-  minimum_key_strength_help =
-    talloc_asprintf (tmp_ctx, "%d or larger", options->minimum_key_strength);
+  minimum_rsa_key_strength_help = talloc_asprintf (
+    tmp_ctx, _ ("%d or larger"), options->minimum_rsa_key_strength);
 
   named_groups_help =
     talloc_asprintf (tmp_ctx,
-                     "Output well-known DH parameters. The available named "
-                     "groups are: %s. (Default: \"ffdhe4096\")",
+                     _ ("Output well-known DH parameters. The available named "
+                        "groups are: %s. (Default: \"ffdhe4096\")"),
                      valid_dh_group_names (tmp_ctx));
 
   options->verbosity = SSCG_DEFAULT;
-  // clang-format off
-  struct poptOption long_options[] = {
-    POPT_AUTOHELP
 
+  // clang-format off
+  struct poptOption verbosity_options[] = {
     {
       "quiet",
       'q',
       POPT_ARG_VAL,
       &options->verbosity,
       SSCG_QUIET,
-       ("Display no output unless there is an error."),
+      _ ("Display no output unless there is an error."),
       NULL
     },
 
@@ -212,26 +234,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       NULL
     },
 
-    {
-      "version",
-      'V',
-      POPT_ARG_NONE,
-      &options->print_version,
-      0,
-      _ ("Display the version number and exit."),
-      NULL
-    },
-
-    {
-      "force",
-      'f',
-      POPT_ARG_NONE,
-      &options->overwrite,
-      0,
-      _ ("Overwrite any pre-existing files in the requested locations"),
-      NULL
-    },
+    POPT_TABLEEND
+  };
 
+  struct poptOption subject_options[] = {
     {
       "lifetime",
       '\0',
@@ -322,36 +328,73 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       _ ("Optional additional valid hostnames for the certificate. "
          "In addition to hostnames, this option also accepts explicit values "
          "supported by RFC 5280 such as "
-         "IP:xxx.xxx.xxx.xxx/yyy.yyy.yyy.yyy "
+         "IP:xxx.xxx.xxx.xxx "
          "May be specified multiple times."),
       _ ("alt.example.com")
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption key_options[] = {
     {
-      "package",
+      "key-type",
       '\0',
       POPT_ARG_STRING,
-      &packagename,
+      &key_type,
       0,
-      _ ("Unused. Retained for compatibility with earlier versions of sscg."),
-      NULL,
+      _ ("Type of key to use for the certificate private keys. (default: "
+         "rsa)"),
+#ifdef HAVE_ML_DSA
+      _ ("{rsa,ecdsa,mldsa}"),
+#else
+      _ ("{rsa,ecdsa}"),
+#endif /* HAVE_ML_DSA */
     },
 
     {
       "key-strength",
       '\0',
       POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
-      &options->key_strength,
+      &options->rsa_key_strength,
       0,
-      _ ("Strength of the certificate private keys in bits."),
-      minimum_key_strength_help },
+      _ ("Strength of the certificate private keys in bits. "
+         "This argument is only valid if --key-type is set to rsa."),
+      minimum_rsa_key_strength_help
+    },
+
+    {
+      "ec-curve",
+      '\0',
+      POPT_ARG_STRING,
+      &ec_curve,
+      0,
+      _ ("EC curve to use for the certificate private keys. "
+         "This argument is only valid if --key-type is set to ecdsa."),
+      _ ("{secp224r1,secp256r1,secp384r1,secp521r1}"),
+    },
+
+#ifdef HAVE_ML_DSA
+    {
+      "mldsa-nist-level",
+      '\0',
+      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
+      &options->mldsa_nist_level,
+      0,
+      _ ("NIST level to use for the ML-DSA key. "
+         "This argument is only valid if --key-type is set to mldsa."),
+      _ ("{2,3,5}"),
+    },
+#endif /* HAVE_ML_DSA */
+
     {
       "hash-alg",
       '\0',
       POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
       &options->hash_alg,
       0,
-      _ ("Hashing algorithm to use for signing."),
+      _ ("Hashing algorithm to use for signing RSA and ECDSA keys. "
+         "This argument is only valid if --key-type is rsa or ecdsa."),
       _ ("{sha256,sha384,sha512}"),
     },
 
@@ -365,6 +408,11 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       _ ("{des-ede3-cbc,aes-256-cbc}"),
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption ca_file_options[] = {
+
     {
       "ca-file",
       '\0',
@@ -440,6 +488,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       NULL
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption crl_file_options[] = {
     {
       "crl-file",
       '\0',
@@ -462,6 +514,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       SSCG_CERT_DEFAULT_MODE_HELP,
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption svc_cert_file_options[] = {
     {
       "cert-file",
       '\0',
@@ -537,6 +593,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       NULL
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption client_cert_file_options[] = {
     {
       "client-file",
       '\0',
@@ -610,6 +670,10 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       NULL
     },
 
+    POPT_TABLEEND
+  };
+
+  struct poptOption dh_file_options[] = {
     {
       "dhparams-file",
       '\0',
@@ -617,17 +681,17 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       &options->dhparams_file,
       0,
       _("A file to contain a set of Diffie-Hellman parameters. "
-        "(Default: \"./dhparams.pem\")"),
+        "(Default: not created)"),
       NULL
     },
 
     {
       "no-dhparams-file",
       '\0',
-      POPT_ARG_NONE,
+      POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN,
       &options->skip_dhparams,
       0,
-      _ ("Do not create the dhparams file"),
+      _ ("Deprecated: Retained for backwards compatibility. To be removed in SSCG 5.0."),
       NULL
     },
 
@@ -644,12 +708,13 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
     {
       "dhparams-prime-len",
       '\0',
-      POPT_ARG_INT | POPT_ARGFLAG_SHOW_DEFAULT,
+      POPT_ARG_INT | POPT_ARGFLAG_DOC_HIDDEN,
       &options->dhparams_prime_len,
       0,
-      _ ("The length of the prime number to generate for dhparams, in bits. "
-         "If set to non-zero, the parameters will be generated rather than "
-         "using a well-known group."),
+      _ ("Deprecated: Retained for backwards compatibility. To be removed in "
+         "SSCG 5.0. The length of the prime number to generate for dhparams, "
+         "in bits. If set to non-zero, the parameters will be generated "
+         "rather than using a well-known group."),
       NULL
     },
 
@@ -665,16 +730,121 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
 
     POPT_TABLEEND
   };
+
+  struct poptOption cli_options[] = {
+    POPT_AUTOHELP
+    {
+      "version",
+      'V',
+      POPT_ARG_NONE,
+      &options->print_version,
+      0,
+      _ ("Display the version number and exit."),
+      NULL
+    },
+
+    {
+      "force",
+      'f',
+      POPT_ARG_NONE,
+      &options->overwrite,
+      0,
+      _ ("Overwrite any pre-existing files in the requested locations"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      verbosity_options,
+      0,
+      _ ("Verbosity options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      subject_options,
+      0,
+      _ ("Certificate Subject options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      key_options,
+      0,
+      _ ("Certificate Key Cryptography Options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      ca_file_options,
+      0,
+      _ ("Certificate Authority File Options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      crl_file_options,
+      0,
+      _ ("Certificate Revocation List File Options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      svc_cert_file_options,
+      0,
+      _ ("Service Certificate File Options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      client_cert_file_options,
+      0,
+      _ ("Client Authentication Certificate File Options"),
+      NULL
+    },
+
+    {
+      NULL,
+      '\0',
+      POPT_ARG_INCLUDE_TABLE,
+      dh_file_options,
+      0,
+      _ ("Diffie-Hellman Parameter File Options"),
+      NULL
+    },
+
+    POPT_TABLEEND
+  };
   // clang-format on
 
-  pc = poptGetContext (argv[0], argc, argv, long_options, 0);
+  pc = poptGetContext (argv[0], argc, argv, cli_options, 0);
   while ((opt = poptGetNextOpt (pc)) != -1)
     {
       switch (opt)
         {
         default:
           fprintf (stderr,
-                   "\nInvalid option %s: %s\n\n",
+                   _ ("\nInvalid option %s: %s\n\n"),
                    poptBadOption (pc, 0),
                    poptStrerror (opt));
           poptPrintUsage (pc, stderr, 0);
@@ -690,7 +860,7 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
       exit (0);
     }
 
-  verbosity = options->verbosity;
+  set_verbosity (options->verbosity);
 
   /* Process the Subject information */
 
@@ -698,7 +868,7 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
     {
       if (strlen (country) != 2)
         {
-          fprintf (stderr, "Country codes must be exactly two letters.\n");
+          fprintf (stderr, _ ("Country codes must be exactly two letters.\n"));
           ret = EINVAL;
           goto done;
         }
@@ -786,10 +956,20 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
     }
   CHECK_MEM (options->hostname);
 
-  if (strnlen (options->hostname, MAXHOSTNAMELEN + 1) > MAXHOSTNAMELEN)
+  if (strnlen (options->hostname, MAX_FQDN_LEN + 1) > MAX_FQDN_LEN)
     {
       fprintf (
-        stderr, "Hostnames may not exceed %d characters\n", MAXHOSTNAMELEN);
+        stderr, _ ("FQDNs may not exceed %d characters\n"), MAX_FQDN_LEN);
+      ret = EINVAL;
+      goto done;
+    }
+
+  if ((strchr (options->hostname, '.') - options->hostname) > MAX_HOST_LEN + 4)
+    {
+      fprintf (stderr,
+               _ ("Hostnames may not exceed %d characters in Subject "
+                  "Alternative Names\n"),
+               MAX_HOST_LEN);
       ret = EINVAL;
       goto done;
     }
@@ -798,30 +978,76 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
        options struct. It's not the most efficient approach, but
        it's only done one time, so there is no sense in optimizing
        it. */
+  size_t i = 0;
   if (alternative_names)
     {
-      size_t i = 0;
       while (alternative_names[i] != NULL)
         {
           options->subject_alt_names = talloc_realloc (
-            options, options->subject_alt_names, char *, i + 2);
+            options, options->subject_alt_names, char *, i + 1);
           CHECK_MEM (options->subject_alt_names);
 
           options->subject_alt_names[i] =
             talloc_strdup (options->subject_alt_names, alternative_names[i]);
           CHECK_MEM (options->subject_alt_names[i]);
-
-          /* Add a NULL terminator to the end */
-          options->subject_alt_names[i + 1] = NULL;
           i++;
         }
     }
 
-  if (options->key_strength < options->minimum_key_strength)
+  /*
+    The hostname must always be listed in SubjectAlternativeNames as well.
+    Note that the realloc also adds an extra entry for the NULL terminator
+  */
+  options->subject_alt_names =
+    talloc_realloc (options, options->subject_alt_names, char *, i + 2);
+  CHECK_MEM (options->subject_alt_names);
+  options->subject_alt_names[i] =
+    talloc_strdup (options->subject_alt_names, options->hostname);
+  CHECK_MEM (options->subject_alt_names[i]);
+  /* Add a NULL terminator to the end */
+  options->subject_alt_names[i + 1] = NULL;
+
+  if (key_type)
     {
-      fprintf (stderr,
-               "Key strength must be at least %d bits.\n",
-               options->minimum_key_strength);
+      talloc_free (options->key_type);
+      options->key_type = talloc_strdup (options, key_type);
+    }
+
+  if (strcmp (options->key_type, "rsa") == 0)
+    {
+      if (options->rsa_key_strength < options->minimum_rsa_key_strength)
+        {
+          fprintf (stderr,
+                   _ ("Key strength must be at least %d bits.\n"),
+                   options->minimum_rsa_key_strength);
+          ret = EINVAL;
+          goto done;
+        }
+    }
+  else if (strcmp (options->key_type, "ecdsa") == 0)
+    {
+      if (ec_curve)
+        {
+          talloc_free (options->ec_curve);
+          options->ec_curve = talloc_strdup (options, ec_curve);
+        }
+    }
+#ifdef HAVE_ML_DSA
+  else if (strcmp (options->key_type, "mldsa") == 0)
+    {
+      if (options->mldsa_nist_level < options->system_security_level)
+        {
+          fprintf (stderr,
+                   _ ("ML-DSA NIST level must be at least %d.\n"),
+                   options->system_security_level);
+          ret = EINVAL;
+          goto done;
+        }
+    }
+#endif /* HAVE_ML_DSA */
+  else
+    {
+      SSCG_ERROR ("Unknown key type: %s\n", options->key_type);
       ret = EINVAL;
       goto done;
     }
@@ -830,27 +1056,42 @@ sscg_handle_arguments (TALLOC_CTX *mem_ctx,
   options->cipher = EVP_get_cipherbyname (options->cipher_alg);
   if (!options->cipher)
     {
-      fprintf (stderr, "Invalid cipher specified: %s\n", options->cipher_alg);
+      fprintf (
+        stderr, _ ("Invalid cipher specified: %s\n"), options->cipher_alg);
       ret = EINVAL;
       goto done;
     }
 
-  /* TODO: restrict this to approved hashes.
-   * For now, we'll only list SHA[256|384|512] in the help */
-  options->hash_fn = EVP_get_digestbyname (options->hash_alg);
+  if (strcmp (options->key_type, "mldsa") == 0)
+    {
+      /* ML-DSA keys do not use a hash algorithm.
+       * If the user specified --hash-alg, it is ignored. */
+      options->hash_alg = NULL;
+      options->hash_fn = NULL;
+    }
+  else
+    {
+      options->hash_fn = EVP_get_digestbyname (options->hash_alg);
+      if (!options->hash_fn
+          || EVP_MD_is_a (options->hash_fn, options->hash_alg) != 1)
+        {
+          /* This check ensures that the passed hash algorithm is both a real
+          * algorithm and that it is permitted by the current system policy. */
+          fprintf (stderr, _ ("Unsupported hashing algorithm.\n"));
+          ret = EINVAL;
+          goto done;
+        }
+    }
+  SSCG_LOG (SSCG_DEBUG,
+            "Hash algorithm: %s, %p\n",
+            options->hash_alg,
+            options->hash_fn);
 
   if (!is_valid_named_group (options->dhparams_group))
     {
-      fprintf (stderr, "Unknown Diffie Hellman finite field group.\n");
+      fprintf (stderr, _ ("Unknown Diffie Hellman finite field group.\n"));
       fprintf (
-        stderr, "Valid groups are: %s.\n", valid_dh_group_names (tmp_ctx));
-      ret = EINVAL;
-      goto done;
-    }
-
-  if (!options->hash_fn)
-    {
-      fprintf (stderr, "Unsupported hashing algorithm.");
+        stderr, _ ("Valid groups are: %s.\n"), valid_dh_group_names (tmp_ctx));
       ret = EINVAL;
       goto done;
     }
@@ -875,23 +1116,23 @@ void
 print_options (struct sscg_options *opts)
 {
   size_t i = 0;
-  fprintf (stdout, "==== Options ====\n");
-  fprintf (stdout, "Certificate lifetime: %d\n", opts->lifetime);
-  fprintf (stdout, "Country: \"%s\"\n", opts->country);
-  fprintf (stdout, "State or Principality: \"%s\"\n", opts->state);
-  fprintf (stdout, "Locality: \"%s\"\n", opts->locality);
-  fprintf (stdout, "Organization: \"%s\"\n", opts->org);
-  fprintf (stdout, "Organizational Unit: \"%s\"\n", opts->org_unit);
-  fprintf (stdout, "Email Address: \"%s\"\n", opts->email);
-  fprintf (stdout, "Hostname: \"%s\"\n", opts->hostname);
+  fprintf (stdout, _ ("==== Options ====\n"));
+  fprintf (stdout, _ ("Certificate lifetime: %d\n"), opts->lifetime);
+  fprintf (stdout, _ ("Country: \"%s\"\n"), opts->country);
+  fprintf (stdout, _ ("State or Principality: \"%s\"\n"), opts->state);
+  fprintf (stdout, _ ("Locality: \"%s\"\n"), opts->locality);
+  fprintf (stdout, _ ("Organization: \"%s\"\n"), opts->org);
+  fprintf (stdout, _ ("Organizational Unit: \"%s\"\n"), opts->org_unit);
+  fprintf (stdout, _ ("Email Address: \"%s\"\n"), opts->email);
+  fprintf (stdout, _ ("Hostname: \"%s\"\n"), opts->hostname);
   if (opts->subject_alt_names)
     {
       for (i = 0; opts->subject_alt_names[i]; i++)
         {
           fprintf (stdout,
-                   "Subject Alternative Name: \"%s\"\n",
+                   _ ("Subject Alternative Name: \"%s\"\n"),
                    opts->subject_alt_names[i]);
         }
     }
-  fprintf (stdout, "=================\n");
+  fprintf (stdout, _ ("=================\n"));
 }
