@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: GPL-3.0-or-later WITH cryptsetup-OpenSSL-exception */
 /*
     This file is part of sscg.
 
@@ -27,7 +28,7 @@
     version.  If you delete this exception statement from all source
     files in the program, then also delete it here.
 
-    Copyright 2019-2023 by Stephen Gallagher <sgallagh@redhat.com>
+    Copyright 2019-2025 by Stephen Gallagher <sgallagh@redhat.com>
 */
 
 
@@ -35,6 +36,12 @@
 #include <path_utils.h>
 #include <string.h>
 #include <talloc.h>
+#include <sys/stat.h>
+
+#include "config.h"
+#ifdef HAVE_GETTEXT
+#include <libintl.h>
+#endif
 
 #include "include/io_utils.h"
 #include "include/key.h"
@@ -231,14 +238,14 @@ sscg_read_pw_file (TALLOC_CTX *mem_ctx, char *path)
   pwdbio = BIO_new_file (path, "r");
   if (pwdbio == NULL)
     {
-      fprintf (stderr, "Can't open file %s\n", path);
+      fprintf (stderr, _ ("Can't open file %s\n"), path);
       return NULL;
     }
 
   /* Read up to one more character than the MAX_PW_LEN */
   for (offset = 0;
-       (i = BIO_read (pwdbio, tpass + offset, MAX_PW_LEN + 1 - offset)) > 0 &&
-       offset < (MAX_PW_LEN + 1);
+       (i = BIO_read (pwdbio, tpass + offset, MAX_PW_LEN + 1 - offset)) > 0
+       && offset < (MAX_PW_LEN + 1);
        offset += i)
     ;
 
@@ -249,7 +256,7 @@ sscg_read_pw_file (TALLOC_CTX *mem_ctx, char *path)
 
   if (i < 0)
     {
-      fprintf (stderr, "Error reading password from BIO\n");
+      fprintf (stderr, _ ("Error reading password from BIO\n"));
       return NULL;
     }
 
@@ -379,7 +386,7 @@ sscg_io_utils_add_output_key (struct sscg_stream **streams,
       stream->passphrase = sscg_read_pw_file (stream, passfile);
       if (!stream->passphrase)
         {
-          fprintf (stderr, "Failed to read passphrase from %s", passfile);
+          fprintf (stderr, _ ("Failed to read passphrase from %s"), passfile);
           ret = EIO;
           goto done;
         }
@@ -450,8 +457,8 @@ io_utils_validate (struct sscg_stream **streams)
         }
 
       /* The CRL file may only contain certificate revocations */
-      if ((stream->filetypes & (1 << SSCG_FILE_TYPE_CRL)) &&
-          (stream->filetypes ^ (1 << SSCG_FILE_TYPE_CRL)))
+      if ((stream->filetypes & (1 << SSCG_FILE_TYPE_CRL))
+          && (stream->filetypes ^ (1 << SSCG_FILE_TYPE_CRL)))
         {
           ret = IO_UTILS_CRL_NON_EXCLUSIVE;
           goto done;
@@ -463,8 +470,8 @@ io_utils_validate (struct sscg_stream **streams)
   /* If the public or private key is present for the service cert, the other
    * must be present also
    */
-  if ((allbits & SSCG_FILE_TYPE_SVC_TYPES) &&
-      ((allbits & SSCG_FILE_TYPE_SVC_TYPES) != SSCG_FILE_TYPE_SVC_TYPES))
+  if ((allbits & SSCG_FILE_TYPE_SVC_TYPES)
+      && ((allbits & SSCG_FILE_TYPE_SVC_TYPES) != SSCG_FILE_TYPE_SVC_TYPES))
     {
       ret = IO_UTILS_SVC_UNMATCHED;
       goto done;
@@ -473,8 +480,9 @@ io_utils_validate (struct sscg_stream **streams)
   /* If the public or private key is present for the client cert, the other
    * must be present also
    */
-  if ((allbits & SSCG_FILE_TYPE_CLIENT_TYPES) &&
-      ((allbits & SSCG_FILE_TYPE_CLIENT_TYPES) != SSCG_FILE_TYPE_CLIENT_TYPES))
+  if ((allbits & SSCG_FILE_TYPE_CLIENT_TYPES)
+      && ((allbits & SSCG_FILE_TYPE_CLIENT_TYPES)
+          != SSCG_FILE_TYPE_CLIENT_TYPES))
     {
       ret = IO_UTILS_CLIENT_UNMATCHED;
       goto done;
@@ -483,8 +491,8 @@ io_utils_validate (struct sscg_stream **streams)
   /* If the private key is present for the CA cert, the public key must be
    * present also
    */
-  if ((allbits & (1 << SSCG_FILE_TYPE_CA_KEY)) &&
-      !(allbits & (1 << SSCG_FILE_TYPE_CA)))
+  if ((allbits & (1 << SSCG_FILE_TYPE_CA_KEY))
+      && !(allbits & (1 << SSCG_FILE_TYPE_CA)))
     {
       ret = IO_UTILS_CA_UNMATCHED;
       goto done;
